"""Script for managing mtg_ssm."""

import csv
import datetime
import os
import shutil

import openpyxl

from mtg_ssm.db import models
from mtg_ssm.mtgjson import downloader
from mtg_ssm.mtgjson import mtgjson
from mtg_ssm.serialization import mtgcsv
from mtg_ssm.serialization import mtgxlsx


def backup_file(filename):
    """Given a filename, backup the file if it exists."""
    if os.path.exists(filename):
        print('Target already exists, making backup.')
        backup_filename = '{0}.bak-{1:%Y-%m-%d_%H-%M-%S}'.format(
            filename, datetime.datetime.now())
        shutil.copyfile(filename, backup_filename)
        print('Backup written to {0}'.format(backup_filename))


def read_mtgjson(db_session, data_path, include_online_only):
    """Read card data from mtgjson to database."""
    print('Attempting to fetch latest mtgjson.')
    new_version = downloader.fetch_mtgjson(data_path)
    if new_version:
        print('Fetched new version of mtgjson.')
    else:
        print('Latest version already present locally.')
    print('Reading mtgjson data from file.')
    mtgdata = downloader.read_mtgjson(data_path)
    print('Updating card information in database.')
    connection = db_session.connection()
    models.Base.metadata.create_all(connection)
    mtgjson.update_models(db_session, mtgdata, include_online_only)
    print('Done updating cards in database.')


def read_csv(db_session, csv_filename):
    """Read card counts from csv file to the database."""
    print('Reading card counts from csv file: {0}'.format(csv_filename))
    with open(csv_filename, 'r') as csv_file:
        reader = csv.DictReader(csv_file)
        mtgcsv.read_row_counts(db_session, reader)
    print('Done reading counts from csv file.')


def write_csv(db_session, csv_filename):
    """Write database contents to a csv file."""
    backup_file(csv_filename)
    print('Writing database contents to csv file: {0}'.format(csv_filename))
    with open(csv_filename, 'w') as csv_file:
        writer = csv.DictWriter(csv_file, mtgcsv.header())
        writer.writeheader()
        for row in mtgcsv.dump_rows(db_session):
            writer.writerow(row)
    print('Done writing database contents to csv file.')


def read_xlsx(db_session, xlsx_filename):
    """Read card counts from xlsx file to the database."""
    print('Reading card counts from xlsx file: {0}'.format(xlsx_filename))
    workbook = openpyxl.load_workbook(filename=xlsx_filename, read_only=True)
    mtgxlsx.read_workbook_counts(db_session, workbook)
    print('Done reading counts from xlsx file.')


def write_xlsx(db_session, xlsx_filename):
    """Write database contents to xlsx file."""
    backup_file(xlsx_filename)
    print('Creating workbook with database contents.')
    workbook = mtgxlsx.dump_workbook(db_session)
    print('Writing workbook to xlsx file: {0}'.format(xlsx_filename))
    workbook.save(xlsx_filename)
    print('Done writing database contents to xlsx file.')



"""Code for handling data from mtgjson."""

import datetime

from mtg_ssm.db import models


def update_models(session, mtg_data, include_online_only):
    """Update database with data from mtgjson (assumes models exist).

    Arguments:
        session: Session, sqlalchemy session to read/write database.
        mtg_data: dict, data from json.load(s) on mtgjson data.
        include_online_only: process online_only sets? if False, skip.
    """
    seen_set_codes = set()
    seen_card_names = set()
    seen_printing_ids = set()
    for set_data in mtg_data.values():
        if not include_online_only and set_data.get('onlineOnly', False):
            continue

        if set_data['code'] in seen_set_codes:
            continue

        card_set = create_set(set_data)
        session.add(card_set)
        seen_set_codes.add(card_set.code)

        for card_data in set_data['cards']:
            if card_data['name'] not in seen_card_names:
                card = create_card(card_data)
                session.add(card)
                seen_card_names.add(card.name)

            if card_data['id'] not in seen_printing_ids:
                printing = create_printing(card_data, set_data['code'])
                session.add(printing)
                seen_printing_ids.add(printing.id)


def create_set(set_data):
    """Given set data, create a CardSet."""
    card_set = models.CardSet(code=set_data['code'])
    card_set.name = set_data['name']
    card_set.block = set_data.get('block')
    card_set.release_date = datetime.datetime.strptime(
        set_data['releaseDate'], '%Y-%m-%d').date()
    card_set.type = set_data['type']
    card_set.online_only = set_data.get('onlineOnly', False)
    return card_set


def create_card(card_data):
    """Given card data, create a Card."""
    card = models.Card(name=card_data['name'])
    card.strict_basic = (card_data.get('supertypes') == ['Basic'])
    return card


def create_printing(card_data, set_code):
    """Given card data, create a CardPrinting."""
    printing = models.CardPrinting(id=card_data['id'])
    printing.card_name = card_data['name']
    printing.set_code = set_code
    printing.set_number = card_data.get('number')
    printing.multiverseid = card_data.get('multiverseid')
    printing.artist = card_data['artist']
    return printing



"""SQLAlchemy models for managing data."""

import enum
import string

import sqlalchemy as sqla
import sqlalchemy.ext.associationproxy as sqlpxy
import sqlalchemy.ext.declarative as sqld
import sqlalchemy.orm as sqlo
import sqlalchemy.orm.collections as sqlc

from mtg_ssm.db import util

VARIANT_CHARS = (string.ascii_lowercase + '★')


class Base(sqld.declarative_base()):
    """Abstract base class for declarative models."""
    __abstract__ = True


class Card(Base):
    """Model for storing card information."""
    __tablename__ = 'cards'
    name = sqla.Column(sqla.Unicode(255), primary_key=True)
    strict_basic = sqla.Column(sqla.Boolean, default=False)

    # Relationships
    printings = sqlo.relationship('CardPrinting')


class CardPrinting(Base):
    """Model for storing information about card printings."""
    __tablename__ = 'printings'
    __table_args__ = (
        sqla.UniqueConstraint(
            'card_name', 'set_code', 'set_number', 'multiverseid',
            name='uix_card_set_number_mvid'),
        {})

    id = sqla.Column(sqla.String(40), primary_key=True)
    card_name = sqla.Column(sqla.ForeignKey('cards.name'))
    set_code = sqla.Column(sqla.ForeignKey('sets.code'))

    set_number = sqla.Column(sqla.Unicode(15), nullable=True)
    multiverseid = sqla.Column(sqla.Integer, nullable=True)

    artist = sqla.Column(sqla.Unicode(255), nullable=False)

    # Properties
    set_integer = sqlo.column_property(
        sqla.cast(
            sqla.func.trim(set_number, VARIANT_CHARS),
            sqla.Integer))
    set_variant = sqlo.column_property(
        sqla.func.trim(set_number, string.digits))

    # Relationships
    card = sqlo.relationship('Card', lazy='joined')
    set = sqlo.relationship('CardSet', lazy='subquery')
    collection_counts = sqlo.relationship(
        'CollectionCount', lazy='subquery',
        collection_class=sqlc.attribute_mapped_collection('key'),
        cascade='all, delete-orphan')
    counts = sqlpxy.association_proxy(
        'collection_counts', 'count',
        creator=lambda k, v: CollectionCount(
            type=getattr(CountTypes, k), count=v))


class CardSet(Base):
    """Model for storing card set information."""
    __tablename__ = 'sets'

    code = sqla.Column(sqla.Unicode(15), primary_key=True)
    name = sqla.Column(sqla.Unicode(255), unique=True, nullable=False)
    block = sqla.Column(sqla.Unicode(255))
    release_date = sqla.Column(sqla.Date)
    type = sqla.Column(sqla.Unicode(255))
    online_only = sqla.Column(sqla.Boolean, default=False)

    # Relationships
    printings = sqlo.relationship(
        'CardPrinting', order_by=(
            CardPrinting.set_integer, CardPrinting.set_variant,
            CardPrinting.multiverseid, CardPrinting.card_name))

    __mapper_args__ = {
        'order_by': release_date,
    }


class CountTypes(enum.IntEnum):
    """Enum for holding card printing types (normal, foil)."""
    copies = 1
    foils = 2


class CollectionCount(Base):
    """Model for storing information about collected printings."""
    __tablename__ = 'collection_counts'

    print_id = sqla.Column(sqla.ForeignKey('printings.id'), primary_key=True)
    type = sqla.Column(util.SqlEnumType(CountTypes), primary_key=True)
    count = sqla.Column(sqla.Integer, nullable=False)

    @property
    def key(self):
        """Model key for use in associationproxy relationships."""
        return self.type.name



"""Methods for managing data in the form of dicts."""

import collections

from mtg_ssm.db import models


def get_printing(
        card_dict, id_to_print, set_name_num_mv_to_prints,
        set_name_mv_to_prints, set_name_num_to_prints):
    """Given a card dict and various indexes, get the matching CardPrinting."""
    mtgjsid = card_dict.get('id')
    set_code = card_dict.get('set')
    name = card_dict.get('name')
    mvid = card_dict.get('multiverseid')
    number = card_dict.get('number')

    if mtgjsid is not None and mtgjsid in id_to_print:
        return id_to_print[mtgjsid]

    if set_code is None or name is None:
        print('Card has no set or card has no name and cannot be found.')
        return None

    print('Could not find card by id, trying (set, name, num, mvid).')
    set_name_num_mv = (set_code, name, number, mvid)
    if set_name_num_mv in set_name_num_mv_to_prints:
        prints = set_name_num_mv_to_prints[set_name_num_mv]
        if len(prints) == 1:
            return prints[0]
        else:
            print('{} entries found.'.format(len(prints)))

    print('trying (set, name, mv)')
    set_name_mv = (set_code, name, mvid)
    if set_name_mv in set_name_mv_to_prints:
        prints = set_name_mv_to_prints[set_name_mv]
        if len(prints) == 1:
            return prints[0]
        else:
            print('{} entries found.'.format(len(prints)))

    print('trying (set, name, number)')
    set_name_num = (set_code, name, number)
    if set_name_num in set_name_num_to_prints:
        prints = set_name_num_to_prints[set_name_num]
        if len(prints) == 1:
            return prints[0]
        else:
            print('{} entries found.'.format(len(prints)))

    print('Warning: Could not find printing for {}'.format(card_dict))
    return None


def load_counts(session, card_dicts):
    """Load counts from dicts of card info/counts into the database."""
    printings = session.query(models.CardPrinting)
    id_to_print = {}
    set_name_num_mv_to_prints = collections.defaultdict(list)
    set_name_mv_to_prints = collections.defaultdict(list)
    set_name_num_to_prints = collections.defaultdict(list)
    for printing in printings:
        set_code = printing.set_code
        name = printing.card_name
        mvid = printing.multiverseid
        num = printing.set_number
        id_to_print[printing.id] = printing
        set_name_num_mv_to_prints[(set_code, name, num, mvid)].append(printing)
        set_name_mv_to_prints[(set_code, name, mvid)].append(printing)
        set_name_num_to_prints[(set_code, name, num)].append(printing)

    for card_dict in card_dicts:
        printing = get_printing(
            card_dict, id_to_print, set_name_num_mv_to_prints,
            set_name_mv_to_prints, set_name_num_to_prints)

        for key in models.CountTypes.__members__.keys():
            count = card_dict.get(key)
            if count is not None:
                printing.counts[key] = count
            elif printing is not None and key in printing.counts:
                del printing.counts[key]



"""Code for handling data in xlsx files."""

import collections
import string

import openpyxl

from mtg_ssm.db import models
from mtg_ssm.serialization import mtgdict


def dump_workbook(session):
    """Return xlsx workbook from the database."""
    workbook = openpyxl.Workbook()

    name_to_prints = collections.defaultdict(list)
    printings = session.query(models.CardPrinting)
    for printing in printings:
        name_to_prints[printing.card_name].append(printing)

    card_sets = session.query(models.CardSet)
    sets_sheet = workbook['Sheet']
    create_sets_sheet(sets_sheet, card_sets)
    for card_set in card_sets:
        cards_sheet = workbook.create_sheet()
        create_cards_sheet(cards_sheet, card_set, name_to_prints)
    return workbook


SETS_SHEET_HEADER = [
    'code', 'name', 'release', 'block', 'type', 'cards', 'unique', 'playsets',
    'count']


def create_sets_sheet(sheet, card_sets):
    """Populate sheet with information about all card sets."""
    sheet.title = 'Sets'
    sheet.append(SETS_SHEET_HEADER)
    sheet.append([
        'Total',
        None,
        None,
        None,
        None,
        '=SUM(F3:F65535)',
        '=SUM(G3:G65535)',
        '=SUM(H3:H65535)',
        '=SUM(I3:I63353)',
    ])
    for card_set in card_sets:
        row = [
            card_set.code,
            card_set.name,
            card_set.release_date,
            card_set.block,
            card_set.type,
            len(card_set.printings),
            '=COUNTIF(\'{}\'!A:A,">0")'.format(card_set.code),
            '=COUNTIF(\'{}\'!A:A,">=4")'.format(card_set.code),
            "=SUM('{}'!A:A)".format(card_set.code),
        ]
        sheet.append(row)

    # Styling
    sheet.freeze_panes = sheet['C3']
    col_width_hidden = [
        ('A', 6, False),
        ('B', 24, False),
        ('C', 12, True),
        ('D', 16, True),
        ('E', 12, True),
        ('F', 6, False),
        ('G', 7, False),
        ('H', 8, False),
        ('I', 7, False),
    ]
    for col, width, hidden in col_width_hidden:
        cdim = sheet.column_dimensions[col]
        cdim.width = width
        cdim.hidden = hidden


def split_into_consecutives(numlist):
    """Split a list of numbers into lists of consecutive groups."""
    if not numlist:
        return []
    numlist.sort()
    retlist = [[]]
    current_group = retlist[0]
    previous = numlist[0] - 1
    for val in numlist:
        if previous + 1 == val:
            current_group.append(val)
        else:
            current_group = [val]
            retlist.append(current_group)
        previous = val
    return retlist


def create_haveref_sum(setcode, rownums):
    """Given a setcode and list of rownumbers, create a sum of have cells."""
    haverefs = []
    for sequence in split_into_consecutives(rownums):
        if len(sequence) == 1:
            haverefs.append("'{0}'!A{1}".format(setcode, sequence[0]))
        else:
            haverefs.append(
                "SUM('{0}'!A{1}:A{2})".format(
                    setcode, sequence[0], sequence[-1]))
    return '+'.join(haverefs)


def get_other_print_references(printing, name_to_prints):
    """Get an xlsx formula to list counts of a card from other sets."""
    if printing.card.strict_basic:
        return None  # Basics are so prolific, they tend to bog things down
    other_prints = (
        p for p in name_to_prints[printing.card_name]
        if p.set_code != printing.set_code)
    setcode_and_release = set()
    setcode_to_rownums = collections.defaultdict(list)
    for other in other_prints:
        other_set = other.set
        setcode_and_release.add((other_set.code, other_set.release_date))
        setcode_to_rownums[other_set.code].append(
            other_set.printings.index(other) + 2)
    if not setcode_to_rownums:
        return None
    setcode_and_release = sorted(setcode_and_release, key=lambda item: item[1])
    set_to_havecellref = collections.OrderedDict()
    for setcode, _ in setcode_and_release:
        rownums = setcode_to_rownums[setcode]
        set_to_havecellref[setcode] = create_haveref_sum(setcode, rownums)
    other_print_references = '=' + '&'.join(
        'IF({1}>0,"{0}: "&{1}&", ","")'.format(k, v)
        for k, v in set_to_havecellref.items())
    return other_print_references


COUNT_KEYS = list(models.CountTypes.__members__.keys())
CARDS_SHEET_HEADER = (
    ['have', 'name', 'id', 'multiverseid', 'number', 'artist'] +
    COUNT_KEYS + ['others'])
COUNT_COLS = [
    string.ascii_uppercase[CARDS_SHEET_HEADER.index(k)] for k in COUNT_KEYS]
HAVE_TMPL = '=' + '+'.join(c + '{0}' for c in COUNT_COLS)
ROW_OFFSET = 2


def create_cards_sheet(sheet, card_set, name_to_prints):
    """Populate sheet with card information from a given set."""
    sheet.title = card_set.code
    sheet.append(CARDS_SHEET_HEADER)
    for printing in card_set.printings:
        rownum = card_set.printings.index(printing) + ROW_OFFSET
        row = [
            HAVE_TMPL.format(rownum),
            printing.card.name,
            printing.id,
            printing.multiverseid,
            printing.set_number,
            printing.artist,
        ]
        for key in COUNT_KEYS:
            row.append(printing.counts.get(key))
        row.append(get_other_print_references(printing, name_to_prints))
        sheet.append(row)

    # Styling
    sheet.freeze_panes = sheet['C2']
    col_width_hidden = [
        ('A', 5, False),
        ('B', 18, False),
        ('C', 5, True),
        ('D', 12, True),
        ('E', 8, True),
        ('F', 20, True),
        ('G', 6, False),
        ('H', 6, False),
        ('I', 10, False),
    ]
    for col, width, hidden in col_width_hidden:
        cdim = sheet.column_dimensions[col]
        cdim.width = width
        cdim.hidden = hidden


def read_workbook_counts(session, workbook):
    """Read mtgxlsx workbook and load counts into the database."""
    card_sets = session.query(models.CardSet)
    set_codes = {s.code for s in card_sets}
    card_dicts = workbook_row_reader(workbook, set_codes)
    mtgdict.load_counts(session, card_dicts)


def workbook_row_reader(workbook, known_sets):
    """Given a workbook, yield card_dicts suitable for mtgdict."""
    for sheet in workbook.worksheets:
        set_code = sheet.title
        if set_code not in known_sets:
            if set_code != 'Sets':
                print('No known set with code "{}", skipping.'.format(set_code))
            continue
        for row_dict in worksheet_row_reader(sheet):
            row_dict['set'] = set_code
            yield row_dict


def worksheet_row_reader(worksheet):
    """Given a worksheet, yield card_dicts."""
    row_iter = iter(worksheet.rows)
    header = [cell.value for cell in next(row_iter)]
    for row in row_iter:
        row_values = [cell.value for cell in row]
        row_dict = dict(zip(header, row_values))
        yield row_dict



"""General helper methods and classes."""

import re

import sqlalchemy.types as sqlt


class SqlEnumType(sqlt.SchemaType, sqlt.TypeDecorator):
    """Type class for storing a python3 enum in SQL.

    Derived from: http://techspot.zzzeek.org/2011/01/14/the-enum-recipe/
    """

    def __init__(self, enum_cls):
        """Given an enum.IntEnum create a sqlalchemy type to store it."""
        super().__init__()
        values = [member.name for member in enum_cls.__members__.values()]
        name = 'ck{}'.format(re.sub(
            '([A-Z])', lambda match: '_' + match.group(1).lower(),
            enum_cls.__name__))
        self.enum = enum_cls
        self.impl = sqlt.Enum(*values, name=name)

    def _set_table(self, column, table):
        self.impl._set_table(column, table)  # pylint: disable=protected-access

    def copy(self):
        return self.__class__(self.enum)

    def process_literal_param(self, value, _dialect):
        return value

    def process_bind_param(self, value, _dialect):
        return None if value is None else value.name

    def process_result_value(self, value, _dialect):
        return None if value is None else getattr(self.enum, value)



#!/usr/bin/env python3
"""Script to manage magic collection spreadsheets."""

import argparse
import os

import sqlalchemy as sqla
import sqlalchemy.orm as sqlo

import mtg_ssm

from mtg_ssm import manager_helper
from mtg_ssm import profiling


MTG_SSM_DATA_PATH = os.path.expanduser(os.path.join('~', '.mtg_ssm'))


def get_parser():
    """Create and return application argument parser."""
    parser = argparse.ArgumentParser(
        description='Magic Collection Spreadsheet Manager')
    parser.add_argument(
        '--version', action='version', version=mtg_ssm.__version__)

    parser.add_argument(
        '--data_path', default=MTG_SSM_DATA_PATH,
        help='Path to mtg_ssm\'s data storage folder. Default={0}'.format(
            MTG_SSM_DATA_PATH))
    parser.add_argument(
        '--include_online_only', default=False, action='store_true',
        help='Include online only sets (e.g. Masters sets) in the database.')
    parser.add_argument(
        '--debug_stats', default=False, action='store_true',
        help='Output additional debugging statistics.')
    parser.add_argument(
        '--db_echo', default=False, action='store_true',
        help='Output sql queries issued against db engine for debugging.')
    parser.add_argument(
        'spreadsheet_file', help='Spreadsheet (xlsx) filename to work with.')

    cmd_subparser = parser.add_subparsers(
        dest='command', help='Operation to perform with the spreadsheet.')
    cmd_subparser.required = True

    cmd_subparser.add_parser(
        'create', help='Create a new, empty spreadsheet.',
        description='Create a new, empty spreadsheet.')

    cmd_subparser.add_parser(
        'update', help='Update card data in spreadsheet.',
        description='Update card data in spreadsheet.')

    export_cmd = cmd_subparser.add_parser(
        'export', help='Export data from spreadsheet in another format.',
        description='Export data from spreadsheet in another format.')
    export_cmd.add_argument(
        '--format', choices=['csv'], default='csv',
        help='Data format for export file.')
    export_cmd.add_argument(
        'export_file', help='Target file for export.')

    import_cmd = cmd_subparser.add_parser(
        'import', help=(
            'Import data to spreadsheet in another format. NOTE: Data in the '
            'spreadsheet will be overwritten (It will be backed up first).'),
        description=(
            'Import data to spreadsheet in another format. NOTE: Data in the '
            'spreadsheet will be overwritten (It will be backed up first).'))
    import_cmd.add_argument(
        '--format', choices=['csv'], default='csv',
        help='Data format of import file.')
    import_cmd.add_argument(
        'import_file', help='Source file for import.')

    return parser


def run_commands(args):
    """Run the requested operations."""
    engine = sqla.create_engine('sqlite://', echo=args.db_echo)
    session_factory = sqlo.sessionmaker(engine)
    session = session_factory()
    try:
        manager_helper.read_mtgjson(
            session, args.data_path, args.include_online_only)
        session.commit()

        if args.command in {'update', 'export'}:
            manager_helper.read_xlsx(session, args.spreadsheet_file)
        session.commit()

        if args.command in {'import'}:
            if args.format == 'csv':
                manager_helper.read_csv(session, args.import_file)
        session.commit()

        if args.command in {'create', 'update'}:
            manager_helper.write_xlsx(session, args.spreadsheet_file)

        if args.command in {'export'}:
            if args.format == 'csv':
                manager_helper.write_csv(session, args.export_file)
    finally:
        session.close()


def main():
    """Process user input and run commands.."""
    parser = get_parser()
    args = parser.parse_args()
    if not os.path.exists(args.data_path):
        os.makedirs(args.data_path)
    elif not os.path.isdir(args.data_path):
        raise Exception(
            'data_path: {} must be a folder'.format(args.data_path))

    profiler = None
    if args.debug_stats:
        profiler = profiling.start()
    try:
        run_commands(args)
    finally:
        if profiler is not None:
            profiling.finish(profiler)


if __name__ == '__main__':
    main()



"""Code for handling data from csv files."""

import sqlalchemy.orm as sqlo

from mtg_ssm.db import models
from mtg_ssm.serialization import mtgdict


def header():
    """Get the header columns for an mtgcsv file."""
    headers = [
        'set',
        'name',
        'number',
        'multiverseid',
        'mtgjid',
    ]
    headers.extend(models.CountTypes.__members__.keys())
    return headers


def dump_rows(session):
    """Yield mtgcsv row dicts from the database."""
    card_sets = session.query(models.CardSet) \
        .options(sqlo.joinedload('printings'))
    for card_set in card_sets:
        for printing in card_set.printings:
            card_info = {
                'set': card_set.code, 'name': printing.card.name,
                'number': printing.set_number,
                'multiverseid': printing.multiverseid,
                'mtgjid': printing.id,
            }
            card_info.update(printing.counts)
            yield card_info


def int_or_none(value):
    """Returns an integer for 0 or non-false values, otherwise None."""
    if value == 0:
        return 0
    elif value:
        return int(value)
    else:
        return None


def process_row_dict(row_dict):
    """Given a row_dict, produce a card_dict suitable for mtgdict."""
    card_dict = {
        'set': row_dict['set'],
        'name': row_dict['name'],
        'number': row_dict['number'] or None,
        'multiverseid': int_or_none(row_dict['multiverseid']),
        'id': row_dict.get('mtgjid'),
    }
    for countype in models.CountTypes.__members__.keys():
        if countype in row_dict:
            card_dict[countype] = int_or_none(row_dict[countype])
    return card_dict


def read_row_counts(session, row_dicts):
    """Read mtgcsv row dicts and load counts into the database."""
    card_dicts = (process_row_dict(row_dict) for row_dict in row_dicts)
    mtgdict.load_counts(session, card_dicts)
